@@ -16,6 +16,7 @@ import Control.Monad.Extra (filterM, unless, unlessM, when, whenJust, whenM)
 import qualified Data.ByteString.Char8 as B
 import Data.Char (isAlphaNum, isDigit)
 import Data.List.Extra
+import Data.List.NonEmpty (NonEmpty(..), nonEmpty)
 import Data.Ord (comparing, Down(Down))
 import Data.Maybe
 import qualified Data.Text as T
@@ -27,11 +28,8 @@ import Network.HTTP.Client (managerResponseTimeout, newManager,
 import Network.HTTP.Client.TLS
 import Network.HTTP.Directory
 import Numeric.Natural
-
 import Options.Applicative (fullDesc, header, progDescDoc)
-
-import Paths_dl_fedora (version)
-
+import qualified Options.Applicative.Help.Pretty as P
 import SimpleCmd (cmd, cmd_, cmdBool, error', grep_, logMsg,
                   pipe_, pipeBool, pipeFile_,
                   warning, (+-+),
@@ -49,6 +47,7 @@ import System.Directory (createDirectory, doesDirectoryExist, doesFileExist,
                          listDirectory, pathIsSymbolicLink, removeFile,
                          renameFile, setCurrentDirectory, withCurrentDirectory,
                          writable)
+import System.Environment (lookupEnv)
 import System.FilePath (dropFileName, isAbsolute, isRelative, makeRelative,
                         joinPath, takeExtension, takeFileName, (</>), (<.>))
 import System.Posix.Files (createSymbolicLink, fileSize, getFileStatus,
@@ -59,10 +58,11 @@ import Text.Read
 import qualified Text.ParserCombinators.ReadP as R
 import qualified Text.ParserCombinators.ReadPrec as RP
 import Text.Regex.Posix
-import qualified Text.PrettyPrint.ANSI.Leijen as P
 
 import DownloadDir
+import Paths_dl_fedora (version)
 
+{- HLINT ignore FedoraEdition "Use camelCase" -}
 data FedoraEdition = Cloud
                    | Container
                    | Everything
@@ -83,6 +83,7 @@ data FedoraEdition = Cloud
                    | Xfce  -- last spin: used below
                    | Silverblue
                    | Kinoite
+                   | COSMIC_Atomic
                    | Onyx
                    | Sericea
                    | IoT
@@ -93,6 +94,7 @@ showEdition rel KDE =
   if fedoraVerOrLater 42 rel
   then "KDE-Desktop"
   else "KDE"
+showEdition _ COSMIC_Atomic = "COSMIC-Atomic"
 showEdition _ KDEMobile = "KDE-Mobile"
 showEdition _ MATE = "MATE_Compiz"
 showEdition _ Miracle = "MiracleWM"
@@ -104,7 +106,7 @@ showShortEdition rel edition =
   if edition == KDE then "KDE" else showEdition rel edition
 
 lowerEdition :: FedoraEdition -> String
-lowerEdition = lower . show
+lowerEdition = replace "_" "-" . lower . show
 
 instance Read FedoraEdition where
   readPrec = do
@@ -138,7 +140,7 @@ allSpins rawhide current rel =
   case rel of
     Rawhide -> allSpins rawhide current $ Fedora rawhide
     Fedora r -> fedoraSpins rel \\ missingEditions r
-    FedoraRespin -> delete KDEMobile $ allSpins rawhide current $ Fedora current
+    FedoraRespin -> allSpins rawhide current (Fedora current) \\ [Cinnamon, KDEMobile, Miracle, Sway]
     FedoraTest -> allSpins rawhide current $ Fedora current -- FIXME use fedora-releases
     FedoraStage -> allSpins rawhide current $ Fedora (current + 1) -- FIXME use fedora-releases
     CS 9 True -> [Cinnamon, KDE, MATE, Xfce] -- FIXME missing MAX, MIN
@@ -161,10 +163,14 @@ allEditions rawhide current rel =
 
 missingEditions :: Natural -> [FedoraEdition]
 missingEditions r =
-  case compare r 41 of
-    GT -> [IoT]
-    EQ -> [COSMIC]
-    LT -> [COSMIC, KDEMobile, Miracle]
+  case compare r 42 of
+    GT -> IoT : [Onyx | r == 43 ] ++
+          -- FIXME missing in rawhide
+          -- https://dl.fedoraproject.org/pub/fedora/linux/development/43/Spins/x86_64/iso/
+          -- https://dl.fedoraproject.org/pub/fedora/linux/development/rawhide/Spins/x86_64/iso/
+          if r > 43 then [I3, SoaS, Xfce] else []
+    EQ -> [IoT, COSMIC_Atomic]
+    LT -> [COSMIC, COSMIC_Atomic]
 
 data RequestEditions = Editions Bool [FedoraEdition]
                      | AllDesktops
@@ -244,22 +250,21 @@ readRelease rawhide current rel =
 data Mode = Check | Local | List | Download Bool -- replace
   deriving Eq
 
+-- FIXME add --editions
+-- FIXME add --releases
 main :: IO ()
 main = do
   let pdoc = Just $ P.vcat
-             [ P.text "Tool for downloading Fedora iso file images.",
-               P.text ("RELEASE = " <> intercalate ", " ["release number", "respin", "rawhide", "test (Beta)", "stage (RC)", "eln", "c9s", "c10s", "c9s-live"]),
-               P.text "EDITION = " <> P.lbrace <> P.align (P.fillCat (P.punctuate P.comma (map (P.text . lowerEdition) [(minBound :: FedoraEdition)..maxBound])) <> P.rbrace) <> P.text " [default: workstation]" ,
-               P.text "",
-               P.text "See <https://github.com/juhp/dl-fedora/#readme>"
+             [ P.pretty "Tool for downloading Fedora iso file images.",
+               P.pretty "EDITION = " <> P.lbrace <> P.align (P.fillCat (P.punctuate P.comma (map (P.pretty . lowerEdition) [(minBound :: FedoraEdition)..maxBound])) <> P.rbrace) <> P.pretty " [default: workstation]" ,
+               P.pretty ("RELEASE = " <> intercalate ", " ["release number", "respin", "rawhide", "test (Beta)", "stage (RC)", "eln", "c9s", "c10s", "c9s-live"]),
+               P.pretty "See <https://github.com/juhp/dl-fedora/#readme>"
              ]
   sysarch <- readArch <$> cmd "rpm" ["--eval", "%{_arch}"]
-  rawhideVersion <- getRawhideVersion
-  currentRelease <- getCurrentFedoraVersion
   home <- getHomeDirectory
   defaultDir <- checkDefaultIsoDir
   simpleCmdArgsWithMods (Just version) (fullDesc <> header "Fedora iso downloader" <> progDescDoc pdoc) $
-    program rawhideVersion
+    program
     <$> switchWith 'g' "gpg-keys" "Import Fedora GPG keys for verifying checksum file"
     <*> checksumOpts
     <*> switchLongWith "debug" "Debug output"
@@ -271,6 +276,7 @@ main = do
     <*> strOptionalWith 'd' "dir" "DIRECTORY" ("Download directory [default:" +-+ prettyDir home defaultDir ++ "]") defaultDir
     <*> switchWith 'n' "dry-run" "Don't actually download anything"
     <*> switchWith 'r' "run" "Boot image in QEMU"
+    <*> optional (strOptionWith 'q' "qemu" "QEMU" "specify QEMU command for --run (overrides $DLFEDORA_QEMU, defaults to qemu-kvm)")
     <*> mirrorOpt
     <*> switchLongWith "dvd" "Download dvd iso instead of boot netinst (for Server, eln, centos)"
     <*> optional (flagLongWith' CSDevelopment "cs-devel" "Use centos-stream development compose" <|>
@@ -278,7 +284,7 @@ main = do
                   flagLongWith' CSProduction "cs-production" "Use centos-stream production compose (default is mirror.stream.centos.org)")
     <*> optional (strOptionLongWith "alt-cs-extra-edition" "('MAX'|'MIN')" "Centos Stream Alternative Live Spin editions (MAX,MIN)")
     <*> (optionWith (eitherReader eitherArch) 'a' "arch" "ARCH" ("Specify arch [default:" +-+ showArch sysarch ++ "]") <|> pure sysarch)
-    <*> (readRelease rawhideVersion currentRelease <$> strArg "RELEASE")
+    <*> strArg "RELEASE"
     <*> (flagLongWith' AllDesktops "all-desktops" "Get all Fedora desktops" <|>
          flagLongWith' AllSpins "all-spins" "Get all Fedora Spins" <|>
          flagLongWith' AllEditions "all-editions" "Get all Fedora editions" <|>
@@ -303,10 +309,14 @@ data Primary = Primary {primaryUrl :: String,
                         primarySize :: Maybe Integer,
                         primaryTime :: Maybe UTCTime}
 
-program :: Natural -> Bool -> CheckSum -> Bool -> Bool -> Mode -> FilePath
-        -> Bool -> Bool -> Mirror -> Bool -> Maybe CentosChannel
-        -> Maybe String -> Arch -> Release -> RequestEditions -> IO ()
-program rawhide gpg checksum debug notimeout mode dlDir dryrun run mirror dvdnet mchannel mcsedition arch tgtrel reqeditions = do
+program :: Bool -> CheckSum -> Bool -> Bool -> Mode -> FilePath
+        -> Bool -> Bool -> Maybe String -> Mirror -> Bool
+        -> Maybe CentosChannel -> Maybe String -> Arch -> String
+        -> RequestEditions -> IO ()
+program gpg checksum debug notimeout mode dlDir dryrun run mqemu mirror dvdnet mchannel mcsedition arch tgtrelstr reqeditions = do
+  rawhide <- getRawhideVersion
+  currentRelease <- getCurrentFedoraVersion
+  let tgtrel = readRelease rawhide currentRelease tgtrelstr
   when (isJust mchannel && not (isCentosStream tgtrel)) $
     error' "channels are only for centos-stream"
   let mirrorUrl =
@@ -349,7 +359,7 @@ program rawhide gpg checksum debug notimeout mode dlDir dryrun run mirror dvdnet
       _ -> error' "--alt-cs-extra-edition is only for CS Alt Live spins"
   current <- getCurrentFedoraVersion
   successes <-
-    mapM (runProgramEdition mgr mirrorUrl showdestdir gpg checksum debug mode dryrun run mirror dvdnet mchannel mcsedition arch tgtrel reqeditions) $
+    mapM (runProgramEdition mgr mirrorUrl showdestdir gpg checksum debug mode dryrun run mqemu mirror dvdnet mchannel mcsedition arch tgtrel reqeditions) $
     if mode == List
     then [Workstation]
     else
@@ -392,10 +402,10 @@ multipleEditions req =
     _ -> True
 
 runProgramEdition :: Manager -> URL -> String -> Bool -> CheckSum -> Bool
-                  -> Mode -> Bool -> Bool -> Mirror -> Bool
-                  -> Maybe CentosChannel -> Maybe String -> Arch -> Release
-                  -> RequestEditions -> FedoraEdition -> IO Bool
-runProgramEdition mgr mirrorUrl showdestdir gpg checksum debug mode dryrun run mirror dvdnet mchannel mcsedition arch tgtrel reqeditions edition =
+                  -> Mode -> Bool -> Bool -> Maybe FilePath -> Mirror -> Bool
+                  -> Maybe CentosChannel -> Maybe String -> Arch
+                  -> Release -> RequestEditions -> FedoraEdition -> IO Bool
+runProgramEdition mgr mirrorUrl showdestdir gpg checksum debug mode dryrun run mqemu mirror dvdnet mchannel mcsedition arch tgtrel reqeditions edition =
   case mode of
     Check -> do
       eres <- findURL True
@@ -427,8 +437,7 @@ runProgramEdition mgr mirrorUrl showdestdir gpg checksum debug mode dryrun run m
       mtarget <- derefSymlink symlink
       whenJust mtarget $ \target -> do
         putStrLn $ "Local:" +-+ target
-        when run $
-          bootImage dryrun target showdestdir
+        when run $ bootImage dryrun target showdestdir mqemu
       return True
     List -> do
       rawhide <- getRawhideVersion
@@ -457,7 +466,7 @@ runProgramEdition mgr mirrorUrl showdestdir gpg checksum debug mode dryrun run m
           unless dryrun $ do
             let localfile = takeFileName fileurl
             updateSymlink localfile symlink removeold
-            when run $ bootImage dryrun localfile showdestdir
+            when run $ bootImage dryrun localfile showdestdir mqemu
           return True
   where
     findURL :: Bool
@@ -527,9 +536,9 @@ runProgramEdition mgr mirrorUrl showdestdir gpg checksum debug mode dryrun run m
             let url = top +/+ path
             when debug $ do
               print url
-              redirs <- retry 3 $ httpRedirects mgr $ replace "https:" "http:" url
-              print redirs
-            ls <- retry 3 $ httpDirectory mgr url
+              -- force redirect
+              httpRedirects mgr (replace "https:" "http:" url) >>= print
+            ls <- retry 2 $ httpDirectory mgr url
             return (url, ls)
 
           findMirror primeUrl path file = do
@@ -714,6 +723,7 @@ runProgramEdition mgr mirrorUrl showdestdir gpg checksum debug mode dryrun run m
     -- https://pagure.io/pungi-fedora/blob/main/f/fedora.conf#_251 kiwibuild
     kiwiSpins :: [FedoraEdition]
     kiwiSpins =
+      [I3 | fedoraVerOrLater 43 tgtrel] ++
       -- https://fedoraproject.org/wiki/Changes/EROFSforLiveMedia
       (if fedoraVerOrLater 42 tgtrel
        then [Budgie, COSMIC, KDE, LXQt, SoaS, Workstation, Xfce]
@@ -724,6 +734,7 @@ runProgramEdition mgr mirrorUrl showdestdir gpg checksum debug mode dryrun run m
 
     editionType :: FedoraEdition -> String
     editionType Server = if dvdnet then "dvd" else "netinst"
+    editionType COSMIC_Atomic = "ostree"
     editionType IoT = "ostree"
     editionType Kinoite = "ostree"
     editionType Onyx = "ostree"
@@ -802,8 +813,15 @@ runProgramEdition mgr mirrorUrl showdestdir gpg checksum debug mode dryrun run m
         (primeSize,primeTime) <- retry 3 $ httpFileSizeTime mgr url
         when debug $ print (primeSize,primeTime,url)
         ok <- checkLocalFileSize filesize checksumfile primeSize primeTime True
-        unless ok $ error' $ "Checksum file filesize mismatch for " ++ checksumfile
-        return ok
+        if ok
+          then return True
+          else
+          if filesize < 900
+          then do
+            warning $ "Renaming unsigned checksum file" +-+ checksumfile
+            renameFile checksumfile $ checksumfile <.> "old"
+            return False
+          else error' $ "Checksum file filesize mismatch for " ++ checksumfile
 
     checkForFedoraKeys :: Natural -> IO Bool
     checkForFedoraKeys n =
@@ -867,7 +885,7 @@ downloadFile dryrun debug done mgr url prime showdestdir = do
       then return Nothing
       else do
       tz <- getCurrentTimeZone
-      putStrLn $ unwords ["downloading", filename, renderTime tz mtime, "to", showdestdir]
+      putStrLn $ unwords ["downloading", filename, renderTime tz mtime, "under", showdestdir]
       doCurl
       return (Just True)
   where
@@ -897,25 +915,42 @@ editionMedia _ = "iso"
 
 liveRespin :: FedoraEdition -> String
 liveRespin Budgie = "Budgie"
+liveRespin COSMIC = "COSMIC"
 liveRespin I3 = "i3"
+liveRespin LXQt = "LXQt"
 liveRespin Miracle = "MiracleWM"
+liveRespin Sway = "Sway"
 liveRespin e = take 4 . upper . showEdition FedoraRespin $ e
 
-bootImage :: Bool -> FilePath -> String -> IO ()
-bootImage dryrun img showdir = do
+bootImage :: Bool -> FilePath -> String -> Maybe String -> IO ()
+bootImage dryrun img showdir mqemu = do
   let fileopts =
         case takeExtension img of
           ".iso" -> ["-boot", "d", "-cdrom"]
           _ -> []
-  mQemu <- findExecutable "qemu-kvm"
-  case mQemu of
-    Just qemu -> do
-      let args = ["-m", "4096", "-rtc", "base=localtime", "-cpu", "host"] ++ fileopts
-      -- replace with showCommandForUser or cmdN when updated
-      putStrLn . unwords $ qemu : map translateInternal (args ++ [showdir </> img])
-      unless dryrun $
-        cmd_ qemu (args ++ [img])
-    Nothing -> error' "Need qemu to run image"
+  qemucmd <-
+    case mqemu of
+      Just q -> return q
+      Nothing -> do
+        menv <- lookupEnv "DLFEDORA_QEMU"
+        return $ fromMaybe "qemu-kvm" menv
+  (c :| args) <-
+    case nonEmpty (words qemucmd) of
+      Nothing -> error' "qemu command cannot be empty"
+      Just (q:|qs) -> do
+        mQemu <- findExecutable q
+        case mQemu of
+          Nothing -> error' $ "No such program:" +-+ q
+          Just qemu ->
+            return $
+            qemu :|
+            if null qs
+            then  ["-m", "4096", "-rtc", "base=localtime", "-cpu", "host"] ++ fileopts
+            else qs
+  -- replace with showCommandForUser or cmdN when updated
+  putStrLn . unwords $ c : args ++ [translateInternal (showdir </> img)]
+  unless dryrun $
+    cmd_ c (args ++ [img])
 
 -- from process System.Process.Posix
 translateInternal :: String -> String
@@ -938,11 +973,11 @@ noTrailingSlash = T.dropWhileEnd (== '/')
 #if !MIN_VERSION_http_directory(0,1,9)
 infixr 5 +/+
 (+/+) :: String -> String -> String
-"" +/+ s = s
-s +/+ "" = s
-s +/+ t | last s == '/' = init s +/+ t
-        | head t == '/' = s +/+ tail t
-s +/+ t = s ++ "/" ++ t
+s +/+ t =
+  case (s,t) of
+    ("",_) -> t
+    (_,"") -> s
+    (_,_) -> dropWhileEnd (== '/') s ++ '/' : dropWhile (== '/') t
 #endif
 
 -- derived from fedora-repoquery Types
